          <section class="m-doc-details" id="{{ data.id }}"><div>
            <h3>
              {{ prefix }}<a href="#{{ data.id }}" class="m-doc-self">{{ data.name }}</a>{% if data.type_link %}: {{ data.type_link }}{% endif %}
              {# the empty line needs to be here to prevent the lines from merging #}

            </h3>
            {% if data.summary %}
            <p>{{ data.summary }}</p>
            {% endif %}
            {% if data.content %}
{{ data.content }}
            {% endif %}
          </div></section>
