          <section class="m-doc-details" id="{{ enum.id }}"><div>
            <h3>
              class {{ prefix }}<a href="#{{ enum.id }}" class="m-doc-self">{{ enum.name }}</a>({{ enum.base_link }})
            </h3>
            {% if enum.summary %}
            <p>{{ enum.summary }}</p>
            {% endif %}
            {% if enum.has_value_details %}
            <table class="m-table m-fullwidth m-flat m-doc">
              <thead><tr><th style="width: 1%">Enumerators</th><th></th></tr></thead>
              <tbody>
                {% for value in enum.values %}
                <tr>
                  <td><a href="#{{ value.id }}" id="{{ value.id }}" class="m-doc-self">{{ value.name }}</a></td>
                  <td>
                  {% if value.content %}
{{ value.content }}
                  {% endif %}
                  </td>
                </tr>
                {% endfor %}
              </tbody>
            </table>
            {% endif %}
            {% if enum.content %}
{{ enum.content }}
            {% endif %}
          </div></section>
