{% extends 'base.html' %}

{% block title %}{% set j = joiner('.') %}{% for name, _ in page.breadcrumb %}{{ j() }}{{ name }}{% endfor %} | {{ super() }}{% endblock %}

{% block main %}
        <h1>
          {%+ for name, target in page.breadcrumb[:-1] %}<span class="m-breadcrumb"><a href="{{ target }}">{{ name }}</a>.<wbr/></span>{% endfor %}{{ page.breadcrumb[-1][0] }}
        </h1>
        {% if page.summary %}
        <p>{{ page.summary }}</p>
        {% endif %}
        {% if page.content %}
{{ page.content }}
        {% endif %}
{% endblock %}
